from typing import List
import hashlib
from collections import defaultdict

from pydantic import ValidationError

from waypoint import WaypointClient
from waypoint.models import StatHistory, GameRound, Stat
from waypoint.data import map_ids
from waypoint.enum import Game, GameVariant


class RoundMapper:
    def __init__(self, client: WaypointClient):
        self.client = client

    @staticmethod
    def _calculate_uid(stat: Stat) -> str:
        ts = stat.DateTime.replace(second=0, microsecond=0).timestamp()
        return hashlib.md5(f"{ts}{stat.MapId}".encode("utf-8")).hexdigest()

    def map_rounds(
        self,
        gamertags: List[str],
        variant: GameVariant = GameVariant.SLAYER,
        game: Game = Game.ALL,
    ) -> List[GameRound]:
        round_map = defaultdict(dict)
        win_map = {}
        map_map = {}
        for page_num in range(1, 11):
            try:
                history: List[StatHistory] = self.client.get_game_history(
                    gamertags,
                    variant,
                    game,
                    page_num,
                )
            except ValidationError:
                continue
            for player_history in history:
                for stat in player_history.Stats:
                    uid = self._calculate_uid(stat)
                    round_map[uid][player_history.Gamertag] = stat
                    map_map[uid] = stat.MapId
                    if stat.Won:
                        win_map[uid] = player_history.Gamertag
        results = []
        for uid, game_round in round_map.items():
            if sorted(list(game_round.keys())) != sorted(gamertags):
                continue
            map_info = map_ids.get(map_map[uid], {})
            results.append(
                GameRound(
                    id=uid,
                    Participants=list(game_round.keys()),
                    Winner=win_map.get(uid),
                    PlayerStats=game_round,
                    Map=map_info.get("map"),
                    Game=map_info.get("game", game),
                    Variant=variant,
                )
            )
        return results



import datetime as dt
from typing import List

from pydantic import BaseModel


class Stat(BaseModel):
    GameId: int
    Score: int
    Kills: int
    Deaths: int
    Assists: int
    Headshots: int
    DateTime: dt.datetime
    MapId: int
    Won: bool
    Medals: int
    GameType: int
    KD: float
    LocalizedMapName: str = None


class StatHistory(BaseModel):
    Gamertag: str
    TotalMultiplayerGamesCompleted: int
    Stats: List[Stat]



from typing import List, Union
import datetime as dt

from requests.packages.urllib3.util.retry import Retry

from waypoint.models import StatHistory
from waypoint.enum import GameVariant, Game
from waypoint.client.base import APIClient, http_method


class WaypointClient(APIClient):
    retry_strategy = Retry(
        total=10,
        backoff_factor=1,
        status_forcelist=[500],
    )

    def __init__(self, auth_cookie: str):
        super().__init__(
            "https://www.halowaypoint.com/en-us/games/halo-the-master-chief-collection/xbox-one/"
        )
        self.session.cookies.set("Auth", auth_cookie)

    @http_method(StatHistory)
    def get_game_history(
        self,
        gamertags: Union[str, List[str]],
        game_variant: GameVariant = GameVariant.SLAYER,
        game: Game = Game.ALL,
        page: int = 1,
    ):
        if isinstance(gamertags, list):
            gamertags = ",".join(gamertags)
        return self.session.get(
            "game-history",
            params={
                "gameVariant": game_variant.value,
                "gamertags": gamertags,
                "page": page,
                "game": game.value,
                "view": "DataOnly",
                "_": round(dt.datetime.now().timestamp() * 1000),
            },
        )



from urllib.parse import urlparse, parse_qs, unquote
from typing import Union

from selenium.webdriver import Chrome, ChromeOptions
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as ec
from webdriver_manager.chrome import ChromeDriverManager


class CookieFetcher:
    oauth_link = (
        "https://login.live.com/oauth20_authorize.srf?client_id=000000004C0BD2F1&scope=xbox.basic+xbox"
        ".offline_access&response_type=code&redirect_uri=https://www.halowaypoint.com/auth/callback&locale"
        "=en-us&display=touch&state=https%3a%2f%2fwww.halowaypoint.com%2fen-us%2fgames%2fhalo-the-master"
        "-chief-collection%2fxbox-one%2fgame-history%3fview%3dDataOnly"
    )
    oauth_dest = unquote(parse_qs(urlparse(oauth_link).query)["state"][0])

    def __init__(self, username: str, password: str):
        self.username = username
        self.password = password

        self.webdriver_service = Service(ChromeDriverManager().install())
        self.webdriver_options = ChromeOptions()
        self.webdriver_options.headless = True
        self.driver: Union[Chrome, None] = None
        self.wait: Union[WebDriverWait, None] = None
        self.get_driver()

    def get_driver(self, timeout: int = 5):
        self.cleanup()
        self.driver = Chrome(
            service=self.webdriver_service,
            options=self.webdriver_options,
        )
        self.wait = WebDriverWait(self.driver, timeout)

    def cleanup(self):
        if self.driver:
            self.driver.quit()

    def perform_auth(self) -> str:
        # Navigate to OAuth link
        self.driver.get(self.oauth_link)

        # Fill out username and click to next step
        username_el = self.driver.find_element(By.NAME, "loginfmt")
        username_el.send_keys(self.username)
        next_btn = self.driver.find_element(By.ID, "idSIButton9")
        next_btn.click()

        # Wait for sign in button to become clickable & fill out password & continue
        self.wait.until(ec.element_to_be_clickable((By.ID, "idSIButton9")))
        password_el = self.driver.find_element(By.NAME, "passwd")
        password_el.send_keys(self.password)
        signin_btn = self.driver.find_element(By.ID, "idSIButton9")
        signin_btn.click()

        # Wait until redirect has finished, then return Auth cookie
        self.wait.until(ec.url_to_be(self.oauth_dest))
        cookie = self.driver.get_cookie("Auth")
        if cookie:
            self.cleanup()
            return cookie["value"]

        # Loop if failure, this is common for this service
        self.get_driver()
        return self.perform_auth()



from enum import Enum, unique


@unique
class GameVariant(str, Enum):
    CTF = "Ctf"
    SLAYER = "Slayer"
    ODDBALL = "Oddball"
    KING_OF_THE_HILL = "Koth"
    JUGGERNAUT = "Juggernaut"
    INFECTION = "Infection"
    FLOOD = "Flood"
    RACE = "Race"
    EXTRACTION = "Extraction"
    DOMINION = "Dominion"
    REGICIDE = "Regicide"
    GRIFBALL = "Grifball"
    RICOCHET = "Ricochet"
    FORGE = "Sandbox"
    VIP = "Vip"
    TERRITORIES = "Territories"
    ASSAULT = "Assault"



import functools
from typing import Callable, Union, TypeVar, Type, List
from urllib.parse import urljoin

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


class BaseURLSession(requests.Session):
    def __init__(self, base_url: str):
        super().__init__()
        self.base_url = base_url

    def request(self, method: str, url: Union[str, bytes], *args, **kwargs):
        return super().request(method, urljoin(self.base_url, url), *args, **kwargs)


class APIClient:
    retry_strategy: Retry = Retry(
        total=10,
        backoff_factor=1,
    )

    def __init__(self, base_url: str):
        adapter = HTTPAdapter(max_retries=self.retry_strategy)
        self.session = BaseURLSession(base_url)
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)


def http_method(model: Type[T] = None):
    def decorator_http_method(
        func: Callable[..., requests.Response]
    ) -> Callable[..., Union[T, List[T], object]]:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Union[T, List[T], object]:
            resp = func(*args, **kwargs)
            resp.raise_for_status()
            data = resp.json()
            if model:
                if isinstance(data, list):
                    return [model(**i) for i in data]
                return model(**data)
            return data

        return wrapper

    return decorator_http_method



from typing import List, Dict, Optional

from pydantic import BaseModel

from waypoint.models.stat import Stat
from waypoint.enum import GameVariant, Game


class GameRound(BaseModel):
    id: str
    Participants: List[str]
    PlayerStats: Dict[str, Stat]
    Winner: Optional[str]
    Map: Optional[str]
    Game: Optional[Game]
    Variant: Optional[GameVariant]



from typing import Dict, TypedDict

from waypoint.enum import Game


class MapEntry(TypedDict):
    game: Game
    map: str


map_ids: Dict[int, MapEntry] = {
    0: {"game": Game.HALO_CE, "map": "Pillar Of Autumn"},
    1: {"game": Game.HALO_CE, "map": "Halo"},
    2: {"game": Game.HALO_CE, "map": "Truth And Reconciliation"},
    3: {"game": Game.HALO_CE, "map": "Silent Cartographer"},
    4: {"game": Game.HALO_CE, "map": "Assault On The Control Room"},
    5: {"game": Game.HALO_CE, "map": "343 Guilty Spark"},
    6: {"game": Game.HALO_CE, "map": "The Library"},
    7: {"game": Game.HALO_CE, "map": "Two Betrayals"},
    8: {"game": Game.HALO_CE, "map": "Keyes"},
    9: {"game": Game.HALO_CE, "map": "The Maw"},
    10: {"game": Game.HALO_CE, "map": "Battle Creek"},
    11: {"game": Game.HALO_CE, "map": "Sidewinder"},
    12: {"game": Game.HALO_CE, "map": "Damnation"},
    13: {"game": Game.HALO_CE, "map": "Rat Race"},
    14: {"game": Game.HALO_CE, "map": "Prisoner"},
    15: {"game": Game.HALO_CE, "map": "Hang Em High"},
    16: {"game": Game.HALO_CE, "map": "Chill Out"},
    17: {"game": Game.HALO_CE, "map": "Derelict"},
    18: {"game": Game.HALO_CE, "map": "Boarding Action"},
    19: {"game": Game.HALO_CE, "map": "Chiron"},
    20: {"game": Game.HALO_CE, "map": "Blood Gulch"},
    21: {"game": Game.HALO_CE, "map": "Wizard"},
    22: {"game": Game.HALO_CE, "map": "Longest"},
    23: {"game": Game.HALO_CE, "map": "Death Island"},
    24: {"game": Game.HALO_CE, "map": "Danger Canyon"},
    25: {"game": Game.HALO_CE, "map": "Infinity"},
    26: {"game": Game.HALO_CE, "map": "Timberland"},
    27: {"game": Game.HALO_CE, "map": "Ice Fields"},
    28: {"game": Game.HALO_CE, "map": "Gephyrophobia"},
    29: {"game": Game.HALO_2, "map": "The Heretic"},
    30: {"game": Game.HALO_2, "map": "The Armory"},
    31: {"game": Game.HALO_2, "map": "Cairo Station"},
    32: {"game": Game.HALO_2, "map": "Outskirts"},
    33: {"game": Game.HALO_2, "map": "Metropolis"},
    34: {"game": Game.HALO_2, "map": "The Arbiter"},
    35: {"game": Game.HALO_2, "map": "The Oracle"},
    36: {"game": Game.HALO_2, "map": "Delta Halo"},
    37: {"game": Game.HALO_2, "map": "Regret"},
    38: {"game": Game.HALO_2, "map": "Sacred Icon"},
    39: {"game": Game.HALO_2, "map": "Quarantine Zone"},
    40: {"game": Game.HALO_2, "map": "Gravemind"},
    41: {"game": Game.HALO_2, "map": "Uprising"},
    42: {"game": Game.HALO_2, "map": "High Charity"},
    43: {"game": Game.HALO_2, "map": "The Great Journey"},
    44: {"game": Game.HALO_2, "map": "Lockout"},
    45: {"game": Game.HALO_2, "map": "Ascension"},
    46: {"game": Game.HALO_2, "map": "Midship"},
    47: {"game": Game.HALO_2, "map": "Ivory Tower"},
    48: {"game": Game.HALO_2, "map": "Beaver Creek"},
    49: {"game": Game.HALO_2, "map": "Burial Mounds"},
    50: {"game": Game.HALO_2, "map": "Colossus"},
    51: {"game": Game.HALO_2, "map": "Zanzibar"},
    52: {"game": Game.HALO_2, "map": "Coagulation"},
    53: {"game": Game.HALO_2, "map": "Headlong"},
    54: {"game": Game.HALO_2, "map": "Waterworks"},
    55: {"game": Game.HALO_2, "map": "Foundation"},
    56: {"game": Game.HALO_2, "map": "Containment"},
    57: {"game": Game.HALO_2, "map": "Warlock"},
    58: {"game": Game.HALO_2, "map": "Sanctuary"},
    59: {"game": Game.HALO_2, "map": "Turf"},
    60: {"game": Game.HALO_2, "map": "Backwash"},
    61: {"game": Game.HALO_2, "map": "Elongation"},
    62: {"game": Game.HALO_2, "map": "Gemini"},
    63: {"game": Game.HALO_2, "map": "Relic"},
    64: {"game": Game.HALO_2, "map": "Terminal"},
    65: {"game": Game.HALO_2, "map": "Desolation"},
    66: {"game": Game.HALO_2, "map": "Tombstone"},
    67: {"game": Game.HALO_2, "map": "District"},
    68: {"game": Game.HALO_2, "map": "Uplift"},
    69: {"game": Game.HALO_3, "map": "Arrival"},
    70: {"game": Game.HALO_3, "map": "Sierra 117"},
    71: {"game": Game.HALO_3, "map": "Crows Nest"},
    72: {"game": Game.HALO_3, "map": "Tsavo Highway"},
    73: {"game": Game.HALO_3, "map": "The Storm"},
    74: {"game": Game.HALO_3, "map": "Floodgate"},
    75: {"game": Game.HALO_3, "map": "The Ark"},
    76: {"game": Game.HALO_3, "map": "The Covenant"},
    77: {"game": Game.HALO_3, "map": "Cortana"},
    78: {"game": Game.HALO_3, "map": "Halo"},
    79: {"game": Game.HALO_3, "map": "Construct"},
    80: {"game": Game.HALO_3, "map": "Epitaph"},
    81: {"game": Game.HALO_3, "map": "Guardian"},
    82: {"game": Game.HALO_3, "map": "High Ground"},
    83: {"game": Game.HALO_3, "map": "Isolation"},
    84: {"game": Game.HALO_3, "map": "Last Resort"},
    85: {"game": Game.HALO_3, "map": "Narrows"},
    86: {"game": Game.HALO_3, "map": "Sandtrap"},
    87: {"game": Game.HALO_3, "map": "Snowbound"},
    88: {"game": Game.HALO_3, "map": "The Pit"},
    89: {"game": Game.HALO_3, "map": "Valhalla"},
    90: {"game": Game.HALO_3, "map": "Foundry"},
    91: {"game": Game.HALO_3, "map": "Rats Nest"},
    92: {"game": Game.HALO_3, "map": "Standoff"},
    93: {"game": Game.HALO_3, "map": "Avalanche"},
    94: {"game": Game.HALO_3, "map": "Blackout"},
    95: {"game": Game.HALO_3, "map": "Ghost Town"},
    96: {"game": Game.HALO_3, "map": "Cold Storage"},
    97: {"game": Game.HALO_3, "map": "Assembly"},
    98: {"game": Game.HALO_3, "map": "Orbital"},
    99: {"game": Game.HALO_3, "map": "Sandbox"},
    100: {"game": Game.HALO_3, "map": "Citadel"},
    101: {"game": Game.HALO_3, "map": "Heretic"},
    102: {"game": Game.HALO_3, "map": "Longshore"},
    103: {"game": Game.HALO_4, "map": "Prologue"},
    104: {"game": Game.HALO_4, "map": "Dawn"},
    105: {"game": Game.HALO_4, "map": "Requiem"},
    106: {"game": Game.HALO_4, "map": "Forerunner"},
    107: {"game": Game.HALO_4, "map": "Infinity"},
    108: {"game": Game.HALO_4, "map": "Reclaimer"},
    109: {"game": Game.HALO_4, "map": "Shutdown"},
    110: {"game": Game.HALO_4, "map": "Composer"},
    111: {"game": Game.HALO_4, "map": "Midnight"},
    112: {"game": Game.HALO_4, "map": "Epilogue"},
    113: {"game": Game.HALO_4, "map": "Multiplayer First Donotuse Or Remove"},
    114: {"game": Game.HALO_4, "map": "Adrift"},
    115: {"game": Game.HALO_4, "map": "Abandon"},
    116: {"game": Game.HALO_4, "map": "Complex"},
    117: {"game": Game.HALO_4, "map": "Exile"},
    118: {"game": Game.HALO_4, "map": "Haven"},
    119: {"game": Game.HALO_4, "map": "Longbow"},
    120: {"game": Game.HALO_4, "map": "Meltdown"},
    121: {"game": Game.HALO_4, "map": "Ragnarok"},
    122: {"game": Game.HALO_4, "map": "Solace"},
    123: {"game": Game.HALO_4, "map": "Vortex"},
    124: {"game": Game.HALO_4, "map": "Ravine"},
    125: {"game": Game.HALO_4, "map": "Impact"},
    126: {"game": Game.HALO_4, "map": "Erosion"},
    127: {"game": Game.HALO_4, "map": "Forge Island"},
    128: {"game": Game.HALO_4, "map": "Wreckage"},
    129: {"game": Game.HALO_4, "map": "Harvest"},
    130: {"game": Game.HALO_4, "map": "Shatter"},
    131: {"game": Game.HALO_4, "map": "Landfall"},
    132: {"game": Game.HALO_4, "map": "Monolith"},
    133: {"game": Game.HALO_4, "map": "Skyline"},
    134: {"game": Game.HALO_4, "map": "Daybreak"},
    135: {"game": Game.HALO_4, "map": "Outcast"},
    136: {"game": Game.HALO_4, "map": "Perdition"},
    137: {"game": Game.HALO_4, "map": "Pitfall"},
    138: {"game": Game.HALO_4, "map": "Vertigo"},
    139: {"game": Game.HALO_4, "map": "Spartan Ops First Donotuse Or Remove"},
    140: {"game": Game.HALO_4, "map": "Ff Chopperbowl"},
    141: {"game": Game.HALO_4, "map": "Ff Sniperalley"},
    142: {"game": Game.HALO_4, "map": "Ff Fortsw"},
    143: {"game": Game.HALO_4, "map": "Ff Temple"},
    144: {"game": Game.HALO_4, "map": "Ff Scurve"},
    145: {"game": Game.HALO_4, "map": "Ff Courtyard"},
    146: {"game": Game.HALO_4, "map": "Ff Complex"},
    147: {"game": Game.HALO_4, "map": "Ff Valhalla"},
    148: {"game": Game.HALO_4, "map": "Ff Factory"},
    149: {"game": Game.HALO_4, "map": "Ff Mezzanie"},
    150: {"game": Game.HALO_4, "map": "Ff Caverns"},
    151: {"game": Game.HALO_4, "map": "Ff Vortex"},
    152: {"game": Game.HALO_4, "map": "Ff Breach"},
    153: {"game": Game.HALO_4, "map": "Ff Hillside"},
    154: {"game": Game.HALO_4, "map": "Ff Engine"},
    155: {"game": Game.HALO_2_ANNIVERSARY, "map": "Lockout"},
    156: {"game": Game.HALO_2_ANNIVERSARY, "map": "Ascension"},
    157: {"game": Game.HALO_2_ANNIVERSARY, "map": "Zanzibar"},
    158: {"game": Game.HALO_2_ANNIVERSARY, "map": "Coagulation"},
    159: {"game": Game.HALO_2_ANNIVERSARY, "map": "Warlock"},
    160: {"game": Game.HALO_2_ANNIVERSARY, "map": "Sanctuary"},
    161: {"game": Game.HALO_2_ANNIVERSARY, "map": "Forge Skybox01"},
    162: {"game": Game.HALO_2_ANNIVERSARY, "map": "Forge Skybox02"},
    163: {"game": Game.HALO_2_ANNIVERSARY, "map": "Forge Skybox03"},
    164: {"game": Game.HALO_3, "map": "Extracredit First Donotuse Or Remove"},
    165: {"game": Game.HALO_3, "map": "Epilogue"},
    177: {"game": Game.HALO_2_ANNIVERSARY, "map": "Relic"},
    178: {"game": Game.HALO_REACH, "map": "Noble Actual"},
    179: {"game": Game.HALO_REACH, "map": "Winter Contingency"},
    180: {"game": Game.HALO_REACH, "map": "Oni Sword Base"},
    181: {"game": Game.HALO_REACH, "map": "Nightfall"},
    182: {"game": Game.HALO_REACH, "map": "Tip Of The Spear"},
    183: {"game": Game.HALO_REACH, "map": "Long Night Of Solace"},
    184: {"game": Game.HALO_REACH, "map": "Exodus"},
    185: {"game": Game.HALO_REACH, "map": "New Alexandria"},
    186: {"game": Game.HALO_REACH, "map": "The Package"},
    187: {"game": Game.HALO_REACH, "map": "The Pillar Of Autumn"},
    188: {"game": Game.HALO_REACH, "map": "The Pillar Of Autumn Credits"},
    189: {"game": Game.HALO_REACH, "map": "Lone Wolf"},
    190: {"game": Game.HALO_REACH, "map": "Boardwalk"},
    191: {"game": Game.HALO_REACH, "map": "Boneyard"},
    192: {"game": Game.HALO_REACH, "map": "Countdown"},
    193: {"game": Game.HALO_REACH, "map": "Powerhouse"},
    194: {"game": Game.HALO_REACH, "map": "Reflection"},
    195: {"game": Game.HALO_REACH, "map": "Spire"},
    196: {"game": Game.HALO_REACH, "map": "Sword Base"},
    197: {"game": Game.HALO_REACH, "map": "Zealot"},
    198: {"game": Game.HALO_REACH, "map": "Anchor 9"},
    199: {"game": Game.HALO_REACH, "map": "Breakpoint"},
    200: {"game": Game.HALO_REACH, "map": "Tempest"},
    201: {"game": Game.HALO_REACH, "map": "Condemned"},
    202: {"game": Game.HALO_REACH, "map": "Highlands"},
    203: {"game": Game.HALO_REACH, "map": "Battle Canyon"},
    204: {"game": Game.HALO_REACH, "map": "Penance"},
    205: {"game": Game.HALO_REACH, "map": "Ridgeline"},
    206: {"game": Game.HALO_REACH, "map": "Solitary"},
    207: {"game": Game.HALO_REACH, "map": "High Noon"},
    208: {"game": Game.HALO_REACH, "map": "Breakneck"},
    209: {"game": Game.HALO_REACH, "map": "Forge World"},
    210: {"game": Game.HALO_REACH, "map": "Beachhead"},
    211: {"game": Game.HALO_REACH, "map": "Corvette"},
    212: {"game": Game.HALO_REACH, "map": "Courtyard"},
    213: {"game": Game.HALO_REACH, "map": "Glacier"},
    214: {"game": Game.HALO_REACH, "map": "Holdout"},
    215: {"game": Game.HALO_REACH, "map": "Outpost"},
    216: {"game": Game.HALO_REACH, "map": "Overlook"},
    217: {"game": Game.HALO_REACH, "map": "Waterfront"},
    218: {"game": Game.HALO_REACH, "map": "Unearthed"},
    219: {"game": Game.HALO_REACH, "map": "Installation 04"},
    220: {"game": Game.HALO_3, "map": "S3D Waterfall"},
    221: {"game": Game.HALO_3, "map": "S3D Edge"},
}



from setuptools import setup, find_packages

with open("requirements.txt", "r") as infile:
    requirements = [i.strip() for i in infile.readlines()]

with open("VERSION", "r") as infile:
    version = infile.read().strip()

with open("README.md", "r") as infile:
    long_description = infile.read()

setup(
    name="waypoint",
    version=version,
    author="Andrew Mickael",
    author_email="andrew.mickael@gmail.com",
    description="Halo Waypoint python client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/amickael/waypoint",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    install_requires=requirements,
    python_requires=">=3.7",
    include_package_data=True,
)



from enum import Enum, unique


@unique
class Game(str, Enum):
    ALL = "All"
    HALO_CE = "HaloCombatEvolved"
    HALO_2 = "Halo2"
    HALO_2_ANNIVERSARY = "Halo2Anniversary"
    HALO_3 = "Halo3"
    HALO_4 = "Halo4"
    HALO_REACH = "HaloReach"
